import 'dart:async';

import 'package:test/test.dart';

import 'package:executor/executor.dart';

void main() {
  group('AsyncTask', () {
    test('does not run after closing', () async {
      final executor = Executor();
      final Future<int> resultFuture =
          executor.scheduleTask(() => Future.microtask(() => 12));
      expect(resultFuture, completion(12));

      await executor.close();
      expect(() => executor.scheduleTask(() => 0), throwsException);
    });

    test('concurrency limit', () async {
      final executor = Executor(concurrency: 2);
      final Completer first = Completer();
      final Completer second = Completer();
      final Completer third = Completer();

      expect(executor.runningCount, 0);
      expect(executor.waitingCount, 0);
      expect(executor.scheduledCount, 0);

      // ignore: unawaited_futures
      executor.scheduleTask(() => first.future);
      // ignore: unawaited_futures
      executor.scheduleTask(() => second.future);
      // ignore: unawaited_futures
      executor.scheduleTask(() => third.future);

      expect(executor.runningCount, 2);
      expect(executor.waitingCount, 1);
      expect(executor.scheduledCount, 3);

      await Future.delayed(Duration(milliseconds: 100));
      expect(executor.runningCount, 2);
      expect(executor.waitingCount, 1);
      expect(executor.scheduledCount, 3);

      first.complete();
      await Future.delayed(Duration(milliseconds: 100));
      expect(executor.runningCount, 2);
      expect(executor.waitingCount, 0);
      expect(executor.scheduledCount, 2);

      second.complete();
      await Future.delayed(Duration(milliseconds: 100));
      expect(executor.runningCount, 1);
      expect(executor.waitingCount, 0);
      expect(executor.scheduledCount, 1);

      third.complete();
      await Future.delayed(Duration(milliseconds: 100));
      expect(executor.runningCount, 0);
      expect(executor.waitingCount, 0);
      expect(executor.scheduledCount, 0);
    });

    test('Exceptions do not block further execution.', () async {
      final executor = Executor(concurrency: 2);
      for (int i = 0; i < 10; i++) {
        // ignore: unawaited_futures
        executor.scheduleTask(() async {
          await Future.delayed(Duration(microseconds: i * 10));
          await Future.microtask(() => throw Exception());
        });
      }
      await executor.join(withWaiting: true);
      await executor.close();
    });

    test('Rate limiting', () async {
      final executor = Executor(concurrency: 2, rate: Rate.perSecond(3));
      final sw = Stopwatch()..start();
      final running = <int>[];
      final times = <int>[];
      for (int i = 0; i < 10; i++) {
        // ignore: unawaited_futures
        executor.scheduleTask(() async {
          running.add(executor.runningCount);
          times.add(sw.elapsedMilliseconds);
          await Future.delayed(Duration(seconds: 1));
        });
      }
      await executor.join(withWaiting: true);
      expect(running, [1, 2, 2, 2, 2, 2, 2, 2, 2, 2]);
      for (int i = 0; i < 10; i += 2) {
        expect(times[i + 1] - times[i], greaterThan(320));
        expect(times[i + 1] - times[i], lessThan(360));
      }
      for (int i = 0; i < 8; i += 2) {
        expect(times[i + 2] - times[i], greaterThan(950));
        expect(times[i + 2] - times[i], lessThan(1050));
      }
    });
  });

  group('StreamTask', () {
    test('completes only after the stream closes', () async {
      final executor = Executor();
      final StreamController controller = StreamController();
      final Stream resultStream =
          executor.scheduleStream(() => controller.stream);
      final Future<List> resultList = resultStream.toList();

      await Future.delayed(Duration(milliseconds: 100));
      expect(executor.runningCount, 1);

      controller.add(1);
      await Future.delayed(Duration(milliseconds: 100));
      expect(executor.runningCount, 1);

      controller.add(2);
      await Future.delayed(Duration(milliseconds: 100));
      expect(executor.runningCount, 1);

      await controller.close();
      await Future.delayed(Duration(milliseconds: 100));
      expect(executor.runningCount, 0);

      expect(resultList, completion([1, 2]));
    });
  });
}
